.status_bar frame {
    /* the status bar is ugly in elementary os otherwise */
    border: 0px;
}

.transparent {
    background: transparent;
}

.copyright {
    font-size: 7pt;
}

.autocompletion label {
    padding: 11px;
}

.autocompletion {
    padding: 0px;
    box-shadow: none;
}

.no-v-padding {
    padding-top: 0;
    padding-bottom: 0;
}

.transformation-draw-area {
    border: 1px solid @borders;
    border-radius: 6px;
    background: @theme_base_color;
}

.multiline > *:first-child {
    /* source view suddenly started showing a white background behind line-numbers */
    /* solution found by furiously trying css rules out in the gtk inspector */
    background: @theme_bg_color;
}

.opaque-text text {
    /* found by roaming through /usr/share/themes,
    and some experimentation in the gnome inspector */
    color: alpha(currentColor, 0.5);
}

/*
@theme_bg_color
@theme_selected_bg_color
@theme_base_color
*/
